import os
import hashlib

from django.conf import settings
from binary_database_files import settings as _settings


def is_fresh(name, content_hash):
    """
    Returns true if the file exists on the local filesystem and matches the
    content in the database. Returns false otherwise.
    """
    if not content_hash:
        return False

    # Check that the actual file exists.
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if not os.path.isfile(fqfn):
        return False

    # Check for cached hash file.
    hash_fn = get_hash_fn(name)
    if os.path.isfile(hash_fn):
        return open(hash_fn).read().strip() == content_hash

    # Otherwise, calculate the hash of the local file.
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if not os.path.isfile(fqfn):
        return False
    local_content_hash = get_file_hash(fqfn)
    return local_content_hash == content_hash


def get_hash_fn(name):
    """
    Returns the filename for the hash file.
    """
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    fqfn_parts = os.path.split(fqfn)
    if not os.path.isdir(fqfn_parts[0]):
        os.makedirs(fqfn_parts[0])
    hash_fn = os.path.join(
        fqfn_parts[0], _settings.DB_FILES_DEFAULT_HASH_FN_TEMPLATE % fqfn_parts[1]
    )
    return hash_fn


def write_file(name, content, overwrite=False):
    """
    Writes the given content to the relative filename under the MEDIA_ROOT.
    """
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if os.path.isfile(fqfn) and not overwrite:
        return
    fqfn_parts = os.path.split(fqfn)
    if not os.path.isdir(fqfn_parts[0]):
        os.makedirs(fqfn_parts[0])
    open(fqfn, "wb").write(content)

    # Cache hash.
    hash_value = get_file_hash(fqfn)
    hash_fn = get_hash_fn(name)
    try:
        value = bytes(hash_value, "utf-8")
    except TypeError:
        value = hash_value
    open(hash_fn, "wb").write(value)

    # Set ownership and permissions.
    uname = getattr(settings, "DATABASE_FILES_USER", None)
    gname = getattr(settings, "DATABASE_FILES_GROUP", None)
    if gname:
        gname = ":" + gname
    if uname:
        os.system('chown -RL %s%s "%s"' % (uname, gname, fqfn_parts[0]))  # noqa: S605

    # Set permissions.
    perms = getattr(settings, "DATABASE_FILES_PERMS", None)
    if perms:
        os.system('chmod -R %s "%s"' % (perms, fqfn_parts[0]))  # noqa: S605


def get_file_hash(fin, force_encoding=None, encoding=None, errors=None, chunk_size=128):
    """
    Iteratively builds a file hash without loading the entire file into memory.
    """

    force_encoding = force_encoding or _settings.DB_FILES_DEFAULT_ENFORCE_ENCODING

    encoding = encoding or _settings.DB_FILES_DEFAULT_ENCODING

    errors = errors or _settings.DB_FILES_DEFAULT_ERROR_METHOD

    if isinstance(fin, str):
        fin = open(fin, "rb")
    h = hashlib.sha512()
    while True:
        text = fin.read(chunk_size)
        if not text:
            break
        if force_encoding:
            if not isinstance(text, str):
                text = str(text, encoding=encoding, errors=errors)
            h.update(text.encode(encoding, errors))
        else:
            h.update(text)
    return h.hexdigest()


def get_text_hash_0004(text):
    """
    Returns the hash of the given text.
    """
    h = hashlib.sha512()
    if not isinstance(text, str):
        text = str(text, encoding="utf-8", errors="replace")
    h.update(text.encode("utf-8", "replace"))
    return h.hexdigest()


def get_text_hash(text, force_encoding=None, encoding=None, errors=None):
    """
    Returns the hash of the given text.
    """

    force_encoding = force_encoding or _settings.DB_FILES_DEFAULT_ENFORCE_ENCODING

    encoding = encoding or _settings.DB_FILES_DEFAULT_ENCODING

    errors = errors or _settings.DB_FILES_DEFAULT_ERROR_METHOD

    h = hashlib.sha512()
    if force_encoding:
        if not isinstance(text, str):
            text = str(text, encoding=encoding, errors=errors)
        h.update(text.encode(encoding, errors))
    else:
        h.update(text)
    return h.hexdigest()



import os

from django.conf import settings
from django.urls import reverse

# If true, when file objects are created, they will be automatically copied
# to the local file system for faster serving.
DB_FILES_AUTO_EXPORT_DB_TO_FS = settings.DB_FILES_AUTO_EXPORT_DB_TO_FS = getattr(
    settings, "DB_FILES_AUTO_EXPORT_DB_TO_FS", True
)


def URL_METHOD_1(name):
    """
    Construct file URL based on media URL.
    """
    return os.path.join(settings.MEDIA_URL, name)


def URL_METHOD_2(name):
    """
    Construct file URL based on configured URL pattern.
    """
    return reverse("database_file", kwargs={"name": name})


URL_METHODS = (
    ("URL_METHOD_1", URL_METHOD_1),
    ("URL_METHOD_2", URL_METHOD_2),
)

DATABASE_FILES_URL_METHOD_NAME = settings.DATABASE_FILES_URL_METHOD_NAME = getattr(
    settings, "DATABASE_FILES_URL_METHOD", "URL_METHOD_1"
)

if callable(settings.DATABASE_FILES_URL_METHOD_NAME):
    method = settings.DATABASE_FILES_URL_METHOD_NAME
else:
    method = dict(URL_METHODS)[settings.DATABASE_FILES_URL_METHOD_NAME]

DATABASE_FILES_URL_METHOD = settings.DATABASE_FILES_URL_METHOD = method

DB_FILES_DEFAULT_ENFORCE_ENCODING = (
    settings.DB_FILES_DEFAULT_ENFORCE_ENCODING
) = getattr(settings, "DB_FILES_DEFAULT_ENFORCE_ENCODING", True)

DB_FILES_DEFAULT_ENCODING = settings.DB_FILES_DEFAULT_ENCODING = getattr(
    settings, "DB_FILES_DEFAULT_ENCODING", "ascii"
)

DB_FILES_DEFAULT_ERROR_METHOD = settings.DB_FILES_DEFAULT_ERROR_METHOD = getattr(
    settings, "DB_FILES_DEFAULT_ERROR_METHOD", "ignore"
)

DB_FILES_DEFAULT_HASH_FN_TEMPLATE = (
    settings.DB_FILES_DEFAULT_HASH_FN_TEMPLATE
) = getattr(settings, "DB_FILES_DEFAULT_HASH_FN_TEMPLATE", "%s.hash")

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



from django.apps import AppConfig


class DatabaseFilesAppConfig(AppConfig):
    """AppConfig to make binary_database_files compatible with app loading"""

    name = "binary_database_files"
    label = "binary_database_files"
    verbose_name = "django-binary-database-files"



import functools
import os
import shutil
import tempfile
import base64
from io import BytesIO
from zipfile import ZipFile

from django.conf import settings
from django.core import files
from django.core.files import File as DjangoFile
from django.core.files.storage import default_storage
from django.core.files.temp import NamedTemporaryFile
from django.core.management import call_command
from django.db import models
from django.test import TestCase, override_settings
from django.core.files.base import ContentFile

from binary_database_files.models import File
from binary_database_files.storage import DatabaseStorage
from binary_database_files.tests.models import Thing
from binary_database_files import utils

DIR = os.path.abspath(os.path.split(__file__)[0])

set_default_file_storage = functools.partial(
    override_settings,
    DEFAULT_FILE_STORAGE="binary_database_files.storage.DatabaseStorage",
)


class DatabaseFilesTestCase(TestCase):
    def setUp(self):
        self.media_dir = os.path.join(DIR, "media/i/special")
        if os.path.isdir(self.media_dir):
            shutil.rmtree(self.media_dir)
        os.makedirs(self.media_dir)

    @set_default_file_storage()
    def test_adding_file(self):

        # Create default thing storing reference to file
        # in the local media directory.
        test_fqfn = os.path.join(self.media_dir, "test.txt")
        open(test_fqfn, "w").write("hello there")
        o1 = o = Thing()
        test_fn = "i/special/test.txt"
        o.upload = test_fn
        o.save()
        obj_id = o.id

        # Confirm thing was saved.
        Thing.objects.update()
        q = Thing.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(q[0].upload.name, test_fn)

        # Confirm the file only exists on the file system
        # and hasn't been loaded into the database.
        q = File.objects.all()
        self.assertEqual(q.count(), 0)

        # Verify that the storage reports that the file exists
        self.assertTrue(o.upload.storage.exists(o.upload.name))

        # Verify we can read the contents of thing.
        o = Thing.objects.get(id=obj_id)
        self.assertEqual(o.upload.read(), b"hello there")

        # Verify that by attempting to read the file, we've automatically
        # loaded it into the database.
        File.objects.update()
        q = File.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(BytesIO(q.first().content).getvalue(), b"hello there")

        # Load a dynamically created file outside /media.
        test_file = files.temp.NamedTemporaryFile(
            suffix=".txt", dir=files.temp.gettempdir()
        )
        data0 = b"1234567890"
        test_file.write(data0)
        test_file.seek(0)
        t = Thing.objects.create(
            upload=files.File(test_file, "temp.txt"),
        )
        self.assertEqual(File.objects.count(), 2)
        t = Thing.objects.get(pk=t.pk)
        self.assertEqual(t.upload.file.size, 10)
        self.assertEqual(t.upload.file.name[-4:], ".txt")
        self.assertEqual(t.upload.file.read(), data0)
        t.upload.delete()
        self.assertEqual(File.objects.count(), 1)

        # Delete file from local filesystem and re-export it from the database.
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)
        # Verify that the storage still reports that the file exists
        self.assertTrue(o1.upload.storage.exists(o1.upload.name))
        o1.upload.read()  # This forces the re-export to the filesystem.
        self.assertEqual(os.path.isfile(test_fqfn), True)

        # This dumps all files to the filesystem.
        File.dump_files()

        # Confirm when delete a file from the database, we also delete it from
        # the filesystem.
        self.assertEqual(default_storage.exists("i/special/test.txt"), True)
        default_storage.delete("i/special/test.txt")
        self.assertEqual(default_storage.exists("i/special/test.txt"), False)
        self.assertEqual(os.path.isfile(test_fqfn), False)

    @set_default_file_storage()
    def test_adding_extzipfile(self):
        # ZipExtFile advertises seek() but can raise UnsupportedOperation
        with ZipFile(os.path.join(DIR, "fixtures/test.zip")) as testzip:
            for filename in testzip.namelist():
                with testzip.open(filename) as testfile:
                    o = Thing()
                    o.upload = DjangoFile(testfile)
                    o.save()
                with testzip.open(filename) as testfile:
                    uploaded = File.objects.get(name="i/special/" + filename)
                    self.assertEqual(uploaded.size, testzip.getinfo(filename).file_size)
                    self.assertEqual(BytesIO(uploaded.content).read(), testfile.read())

    @set_default_file_storage()
    def test_adding_base64_file(self):
        image_content = open(os.path.join(DIR, "fixtures/test_image.png"), "rb").read()
        base64_content = base64.encodebytes(image_content).decode("ascii")
        q = File.objects.count()
        Thing.objects.create(upload=ContentFile(base64_content, name="test_image.png"))
        r = File.objects.count()
        self.assertEqual(r, q + 1)

    def test_hash(self):
        # Create test file.
        image_content = open(os.path.join(DIR, "fixtures/test_image.png"), "rb").read()
        fqfn = os.path.join(self.media_dir, "image.png")
        open(fqfn, "wb").write(image_content)

        # Calculate hash from various sources and confirm they all match.
        expected_hash = (
            "35830221efe45ab0dc3d91ca23c29d2d3c20d00c9afeaa096ab256ec322a7a0b3"
            "293f07a01377e31060e65b4e5f6f8fdb4c0e56bc586bba5a7ab3e6d6d97a192"
        )
        h = utils.get_text_hash(image_content)
        self.assertEqual(h, expected_hash)
        h = utils.get_file_hash(fqfn)
        self.assertEqual(h, expected_hash)
        h = utils.get_text_hash(open(fqfn, "rb").read())
        self.assertEqual(h, expected_hash)

        # Create test file.
        image_content = "aあä"
        fqfn = os.path.join(self.media_dir, "test.txt")
        open(fqfn, "wb").write(image_content.encode("utf-8"))

        expected_hash = (
            "1f40fc92da241694750979ee6cf582f2d5d7d28e18335de05abc54d0560e0f530"
            "2860c652bf08d560252aa5e74210546f369fbbbce8c12cfc7957b2652fe9a75"
        )
        h = utils.get_text_hash(image_content)
        self.assertEqual(h, expected_hash)
        h = utils.get_file_hash(fqfn)
        self.assertEqual(h, expected_hash)
        h = utils.get_text_hash(open(fqfn, "rb").read())
        self.assertEqual(h, expected_hash)

    def test_get_available_name(self):
        storage = DatabaseStorage(location=os.path.join(DIR, "media/custom_location"))
        if os.path.isdir(storage.location):
            shutil.rmtree(storage.location)

        class CustomLocationThing(models.Model):
            upload = models.FileField(storage=storage, max_length=500)

            class Meta:
                managed = False
                db_table = Thing()._meta.db_table

        # Create a file in a subdirectory of the custom location and save it
        test_fn = os.path.join("subdirectory", "test.txt")
        test_fqfn = os.path.join(storage.location, "subdirectory", "test.txt")
        with NamedTemporaryFile() as t:
            t.write(b"hello there")
            o = CustomLocationThing()
            o.upload = files.File(t, name=test_fn)
            o.save()

        # Verify we can read the contents of thing.
        o = CustomLocationThing.objects.first()
        self.assertEqual(o.upload.read(), b"hello there")
        self.assertEqual(File.objects.count(), 1)
        self.assertEqual(
            File.objects.first().name,
            os.path.join("custom_location", "subdirectory", "test.txt"),
        )

        # Delete file from local filesystem
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)

        # Create another (different) file in the same subdirectory of the custom location and save it
        with NamedTemporaryFile() as t:
            t.write(b"goodbye")
            o2 = o = CustomLocationThing()
            o.upload = files.File(t, name=test_fn)
            o.save()

        # Verify we can read the contents of thing.
        o = CustomLocationThing.objects.get(pk=o2.pk)
        self.assertEqual(o.upload.read(), b"goodbye")
        self.assertEqual(File.objects.count(), 2)
        # 'custom_location/subdirectory/test.txt' is already taken, so the second file
        # should have a different name
        self.assertNotEqual(
            o.upload.name, os.path.join("custom_location", "subdirectory", "test.txt")
        )
        self.assertTrue(
            o.upload.name.startswith(
                os.path.join("custom_location", "subdirectory", "test_")
            )
        )

    def test_custom_location(self):
        storage = DatabaseStorage(location=os.path.join(DIR, "media/custom_location"))
        if os.path.isdir(storage.location):
            shutil.rmtree(storage.location)

        class CustomLocationThing(models.Model):
            upload = models.FileField(storage=storage, max_length=500)

            class Meta:
                managed = False
                db_table = Thing()._meta.db_table

        # Create a file in a subdirectory of the custom location and save it
        test_fn = os.path.join("subdirectory", "test.txt")
        test_fqfn = os.path.join(storage.location, "subdirectory", "test.txt")
        os.makedirs(os.path.join(storage.location, "subdirectory"))
        open(test_fqfn, "w").write("hello there")
        o1 = o = CustomLocationThing()
        o.upload = test_fn
        o.save()

        # Confirm thing was saved.
        q = CustomLocationThing.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(q[0].upload.name, test_fn)

        # Confirm the file only exists on the file system
        # and hasn't been loaded into the database.
        q = File.objects.all()
        self.assertEqual(q.count(), 0)

        # Verify we can read the contents of thing.
        o = CustomLocationThing.objects.first()
        self.assertEqual(o.upload.read(), b"hello there")

        # Verify that by attempting to read the file, we've automatically
        # loaded it into the database.
        self.assertEqual(q.count(), 1)
        self.assertEqual(BytesIO(q.first().content).getvalue(), b"hello there")

        # Verify that the storage reports that the file exists
        self.assertTrue(o.upload.storage.exists(o.upload.name))

        # Delete file from local filesystem and re-export it from the database.
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)
        # Verify that the storage still reports that the file exists
        self.assertTrue(o1.upload.storage.exists(o1.upload.name))
        o1.upload.read()  # This forces the re-export to the filesystem.
        self.assertEqual(os.path.isfile(test_fqfn), True)

        # Verify that deleting the file from the storage removes both the database and the local copy
        o.upload.delete()
        self.assertEqual(os.path.isfile(test_fqfn), False)
        self.assertEqual(q.count(), 0)
        self.assertFalse(o1.upload.storage.exists(o1.upload.name))

    def test_duplicate_name_in_different_locations(self):
        storage1 = DatabaseStorage(location=os.path.join(DIR, "media/location1"))
        if os.path.isdir(storage1.location):
            shutil.rmtree(storage1.location)

        storage2 = DatabaseStorage(location=os.path.join(DIR, "media/location2"))
        if os.path.isdir(storage2.location):
            shutil.rmtree(storage2.location)

        def get_name(instance, filename):
            return "dummy.txt"

        class Location1Thing(models.Model):
            upload = models.FileField(
                storage=storage1, upload_to=get_name, max_length=500
            )

            class Meta:
                managed = False
                db_table = Thing()._meta.db_table

        class Location2Thing(models.Model):
            upload = models.FileField(
                storage=storage2, upload_to=get_name, max_length=500
            )

            class Meta:
                managed = False
                db_table = Thing()._meta.db_table

        tmpdir = tempfile.mkdtemp(dir=os.path.join(settings.PROJECT_DIR, "media"))
        data1 = b"11111111"
        open(os.path.join(tmpdir, "dummy.txt"), "wb").write(data1)
        upload = files.File(open(os.path.join(tmpdir, "dummy.txt"), "rb"))
        t1 = Location1Thing(upload=upload)
        t1.save()
        self.assertTrue(t1.upload.storage.exists(t1.upload.name))
        os.remove(t1.upload.path)
        self.assertTrue(t1.upload.storage.exists(t1.upload.name))
        self.assertRaises(NotImplementedError, lambda t1: t1.upload.path, t1)
        data2 = b"22222222"
        open(os.path.join(tmpdir, "dummy.txt"), "wb").write(data2)
        t2 = Location2Thing.objects.create(
            upload=files.File(open(os.path.join(tmpdir, "dummy.txt"), "rb"))
        )
        os.remove(t2.upload.path)
        self.assertTrue(t2.upload.storage.exists(t2.upload.name))
        self.assertRaises(NotImplementedError, lambda t2: t2.upload.path, t2)
        self.assertEqual(File.objects.count(), 2)
        self.assertEqual(Location2Thing.objects.get(pk=t2.pk).upload.file.read(), data2)
        self.assertEqual(Location1Thing.objects.get(pk=t1.pk).upload.file.read(), data1)
        shutil.rmtree(tmpdir)

    def test_custom_upload_to(self):
        storage = DatabaseStorage(location=os.path.join(DIR, "media/custom_location"))
        if os.path.isdir(storage.location):
            shutil.rmtree(storage.location)

        def get_name(instance, filename):
            return "dummy.txt"

        class CustomUploadToThing(models.Model):
            upload = models.FileField(
                storage=storage, upload_to=get_name, max_length=500
            )

            class Meta:
                managed = False
                db_table = Thing()._meta.db_table

        # Create a file in a subdirectory of the custom location and save it
        test_fn = os.path.join("subdirectory", "test.txt")
        test_fqfn = os.path.join(storage.location, "subdirectory", "test.txt")
        os.makedirs(os.path.join(storage.location, "subdirectory"))
        open(test_fqfn, "w").write("hello there")
        o1 = o = CustomUploadToThing()
        o.upload = test_fn
        o.save()

        # Confirm thing was saved.
        q = CustomUploadToThing.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(q[0].upload.name, test_fn)

        # Confirm the file only exists on the file system
        # and hasn't been loaded into the database.
        q = File.objects.all()
        self.assertEqual(q.count(), 0)

        # Verify we can read the contents of thing.
        o = CustomUploadToThing.objects.first()
        self.assertEqual(o.upload.read(), b"hello there")

        # Verify that by attempting to read the file, we've automatically
        # loaded it into the database.
        self.assertEqual(q.count(), 1)
        self.assertEqual(BytesIO(q.first().content).getvalue(), b"hello there")

        # Verify that the storage reports that the file exists
        self.assertTrue(o.upload.storage.exists(o.upload.name))

        # Delete file from local filesystem and re-export it from the database.
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)
        # Verify that the storage still reports that the file exists
        self.assertTrue(o1.upload.storage.exists(o1.upload.name))
        o1.upload.read()  # This forces the re-export to the filesystem.
        self.assertEqual(os.path.isfile(test_fqfn), True)

    def test_reading_file(self):
        call_command("loaddata", "test_files.json")
        self.assertEqual(File.objects.count(), 1)
        response = self.client.get("/files/1.txt")
        if hasattr(response, "streaming_content"):
            content = list(response.streaming_content)[0]
        else:
            content = response.content
        self.assertEqual(content, b"1234567890")
        self.assertEqual(response["content-type"], "text/plain")
        self.assertEqual(response["content-length"], "10")

    def test_serve_file_from_database(self):
        call_command("loaddata", "test_files.json")
        self.assertEqual(File.objects.count(), 1)
        test_fqfn = os.path.join(DIR, "media", "1.txt")
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)
        response = self.client.get("/files/1.txt")
        if hasattr(response, "streaming_content"):
            content = list(response.streaming_content)[0]
        else:
            content = response.content
        self.assertEqual(content, b"1234567890")
        self.assertEqual(response["content-type"], "text/plain")
        self.assertEqual(response["content-length"], "10")



import os

PROJECT_DIR = os.path.dirname(__file__)

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3"}}

ROOT_URLCONF = "binary_database_files.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "binary_database_files",
    "binary_database_files.tests",
]

MEDIA_ROOT = os.path.join(PROJECT_DIR, "media")

USE_TZ = True

SECRET_KEY = "secret"  # noqa: S105

AUTH_USER_MODEL = "auth.User"

SITE_ID = 1

BASE_SECURE_URL = "https://localhost"

BASE_URL = "http://localhost"

MIDDLEWARE_CLASSES = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    # 'django.middleware.transaction.TransactionMiddleware',
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.locale.LocaleMiddleware",
)



import mimetypes

from django.conf import settings
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_control
from django.views.static import serve as django_serve

from binary_database_files.models import File
from binary_database_files import settings as _settings


@cache_control(max_age=86400)
def serve(request, name):
    """
    Retrieves the file from the database.
    """
    f = get_object_or_404(File, name=name)
    if _settings.DB_FILES_AUTO_EXPORT_DB_TO_FS:
        f.dump()
    mimetype = mimetypes.guess_type(name)[0] or "application/octet-stream"
    # Cast to bytes to work around https://code.djangoproject.com/ticket/30294
    response = HttpResponse(bytes(f.content), content_type=mimetype)
    response["Content-Length"] = f.size
    return response


def serve_mixed(request, *args, **kwargs):
    """
    First attempts to serve the file from the filesystem,
    then tries the database.
    """
    name = kwargs.get("name") or kwargs.get("path")
    document_root = kwargs.get("document_root")
    document_root = document_root or settings.MEDIA_ROOT
    try:
        # First attempt to serve from filesystem.
        return django_serve(request, name, document_root)
    except Http404:
        # Then try serving from database.
        return serve(request, name)



#!/usr/bin/env python
import io
import os
from setuptools import setup, find_packages

import binary_database_files

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
DESCRIPTION = (
    "A storage system for Django that stores uploaded files in both the "
    "database and file system."
    "Upstream: https://pypi.org/project/django-binary-database-files/ ."
    "+Python 3.1x."
)


def get_reqs(*fns):
    lst = []
    for fn in fns:
        for package in open(os.path.join(CURRENT_DIR, fn)).readlines():
            package = package.strip()
            if not package:
                continue
            lst.append(package.strip())
    return lst


try:
    with io.open(os.path.join(CURRENT_DIR, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setup(
    name="django-binary-database-files-dimaqq-py31x",
    version=binary_database_files.__version__,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Roger Hunwicks",
    author_email="roger@tonic-solutions.com",
    url="https://github.com/kimetrica/django-binary-database-files/",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 6 - Mature",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
    ],
    install_requires=get_reqs(
        "pip-requirements.txt",
    ),
    tests_require=get_reqs("pip-requirements-test.txt"),
    python_requires=">=3.6,<4",
)



from django.conf import settings
from django.core.files.storage import default_storage
from django.core.management.base import BaseCommand
from django.db.models import FileField, ImageField
from django.apps import apps

from binary_database_files.models import File


class Command(BaseCommand):
    args = ""
    help = (
        "Deletes all files in the database that are not referenced by "
        "any model fields."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--dryrun",
            action="store_true",
            dest="dryrun",
            default=False,
            help=(
                "If given, only displays the names of orphaned files "
                "and does not delete them."
            ),
        )
        parser.add_argument(
            "--filenames",
            default="",
            help="If given, only files with these names will be checked",
        )

    def handle(self, *args, **options):
        tmp_debug = settings.DEBUG
        settings.DEBUG = False
        names = set()
        dryrun = options["dryrun"]
        filenames = {_.strip() for _ in options["filenames"].split(",") if _.strip()}
        try:
            for model in apps.get_models():
                print("Checking model %s..." % (model,))
                for field in model._meta.fields:
                    if not isinstance(field, (FileField, ImageField)):
                        continue
                    # Ignore records with null or empty string values.
                    q = {"%s__isnull" % field.name: False}
                    xq = {field.name: ""}
                    subq = model.objects.filter(**q).exclude(**xq)
                    subq_total = subq.count()
                    subq_i = 0
                    for row in subq.iterator():
                        subq_i += 1
                        if subq_i == 1 or not subq_i % 100:
                            print("%i of %i" % (subq_i, subq_total))
                        f = getattr(row, field.name)
                        if f is None:
                            continue
                        if not f.name:
                            continue
                        names.add(f.name)

            # Find all database files with names not in our list.
            print("Finding orphaned files...")
            orphan_files = File.objects.exclude(name__in=names)
            if filenames:
                orphan_files = orphan_files.filter(name__in=filenames)
            orphan_files = orphan_files.only("name", "size")
            total_bytes = 0
            orphan_total = orphan_files.count()
            orphan_i = 0
            print("Deleting %i orphaned files..." % (orphan_total,))
            for f in orphan_files.iterator():
                orphan_i += 1
                if orphan_i == 1 or not orphan_i % 100:
                    print("%i of %i" % (orphan_i, orphan_total))
                total_bytes += f.size
                if dryrun:
                    print("File %s is orphaned." % (f.name,))
                else:
                    print("Deleting orphan file %s..." % (f.name,))
                    default_storage.delete(f.name)
            print("%i total bytes in orphan files." % total_bytes)
        finally:
            settings.DEBUG = tmp_debug
